// Functions for IntMap2
// Copyright (c) Steven Lawrance 1995, 1998

#include <graphics.h>
#include <stdio.h>
#include <stdlib.h>
#include "xmsmem.h"

// Windows Bitmap Support
typedef struct tagRGBTRIPLE
{
	 BYTE    rgbtBlue;
    BYTE    rgbtGreen;
    BYTE    rgbtRed;
} RGBTRIPLE;
typedef RGBTRIPLE FAR* LPRGBTRIPLE;

typedef struct tagRGBQUAD
{
	 BYTE    rgbBlue;
	 BYTE    rgbGreen;
	 BYTE    rgbRed;
	 BYTE    rgbReserved;
} RGBQUAD;
typedef RGBQUAD FAR* LPRGBQUAD;

typedef struct tagBITMAPINFOHEADER
{
	 DWORD   biSize;
	 LONG    biWidth;
	 LONG    biHeight;
	 WORD    biPlanes;
	 WORD    biBitCount;
	 DWORD   biCompression;
	 DWORD   biSizeImage;
	 LONG    biXPelsPerMeter;
	 LONG    biYPelsPerMeter;
	 DWORD   biClrUsed;
	 DWORD   biClrImportant;
} BITMAPINFOHEADER;
typedef BITMAPINFOHEADER FAR* LPBITMAPINFOHEADER;

/* constants for the biCompression field */
#define BI_RGB      0L
#define BI_RLE8     1L
#define BI_RLE4     2L

typedef struct tagBITMAPINFO
{
	 BITMAPINFOHEADER bmiHeader;
	 RGBQUAD          bmiColors[1];
} BITMAPINFO;
typedef BITMAPINFO FAR* LPBITMAPINFO;

typedef struct tagBITMAPFILEHEADER
{
	 UINT    bfType;
	 DWORD   bfSize;
	 UINT    bfReserved1;
	 UINT    bfReserved2;
	 DWORD   bfOffBits;
} BITMAPFILEHEADER;
typedef BITMAPFILEHEADER FAR* LPBITMAPFILEHEADER;

#define LEFT    TRUE
#define RIGHT   FALSE

#define INTMAP_PRESERVESYSTEMPALETTE  0x02
#define INTMAP_PRESERVEENTIREPALETTE  0x03

typedef const char far *LPCSTR;
typedef char far       *LPSTR;

extern class IntMap {
  public:
	 DWORD maxConventionalSize;
	 BOOL criticalSection;
	 BYTE backgroundColor;

	 BOOL LoadIntMap(LPSTR FileName);
	 BOOL PasteIntMap(UINT x, UINT y, UINT w, UINT h);
	 BOOL EraseRegion(UINT x, UINT y, UINT w, UINT h);
	 BOOL ClippedEraseRegion(UINT x, UINT y, UINT w, UINT h, UINT cx, UINT cy, UINT cw, UINT ch);
	 BOOL GetScreenIntMap(UINT sx, UINT sy, UINT ix, UINT iy, UINT iw, UINT ih);
	 BOOL GetIntMapFromScreen(UINT sx, UINT sy, UINT ix, UINT iy, UINT iw, UINT ih);
	 static BOOL CopyIntMap(UINT x1, UINT y1, UINT w, UINT h, UINT x2, UINT y2, IntMap NewIntMap, IntMap OldIntMap);
	 static BOOL IntersectRect(UINT x1, UINT y1, UINT w1, UINT h1, UINT x2, UINT y2, UINT w2, UINT h2, UINT *xi, UINT *yi, UINT *wi, UINT *hi);
	 BOOL MoveIntMap(UINT x1, UINT y1, UINT w1, UINT h1, UINT x2, UINT y2, UINT w2, UINT h2);
	 BOOL RotateIntMap(BOOL direction);
	 UINT GetWidth(void);
	 UINT GetHeight(void);

  private:
	 UINT bgX, bgY, bgW, bgH;
	 UINT width, height;
	 HGLOBAL hIntMapData;
	 HGLOBAL hBackgroundData;
	 HGLOBAL hPalette;
	 LPBYTE lpIntMapData;
	 LPBYTE lpBackgroundData;
	 BOOL criticalSectionOn;

	 BYTE NearestColor(COLORREF target, BYTE palEntry, LPRGBTRIPLE lpPalette, LPBYTE lpMergeDecoder, BYTE nEntries);
};

BOOL IntMap::LoadIntMap(LPCSTR FileName) {
  FILE hFile;
  HGLOBAL hNewBackgroundData;
  UINT x, y;

  // Get rid of any previous allocation
  if (hIntMapData) {
	 if (criticalSectionOn) {
		criticalSectionOn = FALSE;
		GlobalArrayEnd(hIntMapData);
		GlobalArrayEnd(hBackgroundData);
	 }
	 GlobalFree(hIntMapData);
	 hIntMapData = NULL;
  }

  // Open IntMap file and read the number of elements
  hFile = fopen(FileName, "r");
  if (!hFile)
	 return FALSE;
  scanf("%u %u", &width, &height);
  width++;
  height++;

  // Resize the background, not caring about the data inside
  hNewBackgroundData = GlobalReAlloc(hBackgroundData, width * height, GMEM_MOVEABLE | GMEM_ZEROINIT);
  if (!hNewBackgroundData) {
	 fclose(hFile);
	 return FALSE;
  }
  hBackgroundData = hNewBackgroundData;

  // Allocate new memory
  hIntMapData = GlobalAlloc(GHND, width * height);
  if (!hIntMapData) {
	 fclose(hFile);
	 return FALSE;
  }
  if (!GlobalArrayStart(hIntMapData, width, height, (maxConventionalSize / 2) / width)) {
	 GlobalFree(hIntMapData);
	 fclose(hFile);
	 hIntMapData = NULL;
	 return FALSE;
  }

  // Read the IntMap into memory (change the background color, if
  // detected to the current background color index)
  for (y = 0; y < height; y++) {
	 lpIntMapData = (LPBYTE) GlobalArrayElement(hIntMapData, y);
	 for (x = 0; x < width; x++) {
		scanf("%hu", &lpIntMapData[x]);
		if (lpIntMapData[x] == 20) // 20 is the IntMap1 background color
		  lpIntMapData[x] = backgroundColor;
	 }
  }

  // Close the file
  fclose(hFile);

  // If the programmer wanted a critical section, then don't end the global array
  if (criticalSection) {
	 criticalSectionOn = TRUE;
	 GlobalArrayStart(hBackgroundData, width, height, (maxConventionalSize / 2) / width);
  } else {
	 criticalSectionOn = FALSE;
	 GlobalArrayEnd(hIntMapData);
  }
}

BOOL IntMap::LoadIntMapEx(LPCSTR FileName, UINT x, UINT y, UINT w, UINT h, UINT ix, UINT iy, WORD wFlags) {
  HFILE hFile;
  LONG fileLength;

  // Load entire file into RAM
  hFile = fopen(FileName, "r");
  if (!hFile)
	 return FALSE;

  // Get the length of the file
  fseek(hFile, 0, SEEK_END);
  fileLength = ftell(hFile);

  if (fileLength > 0) {
	 HGLOBAL hFileData = GlobalAlloc(GHND, fileLength);
	 LPVOID curBlock;
	 DWORD bytesLeft = fileLength;
	 DWORD block = 0;
	 BOOL retVal;

	 if (!hFileData) {
		fclose(hFile);
		return FALSE;
	 }

	 // Read the file into the memory
	 fseek(hFile, 0, SEEK_SET);
	 while ((bytesLeft / 32767) > 0) {
		curBlock = GlobalLockEx(hFileData, TRUE, block, 32767);
		fread(curBlock, 32767, 1, hFile);
		GlobalUnlock(hFileData);
		block += 32767;
		bytesLeft -= 32767;
	 }
	 if ((bytesLeft) > 0) {
		curBlock = GlobalLockEx(hFileData, TRUE, block, bytesLeft, wFlags);
		fread(curBlock, bytesLeft, 1, hFile);
		GlobalUnlock(hFileData);
	 }
	 fclose(hFile);

	 // Now that it is in memory, call upon LoadIntMapFromMemory to
	 // read it :-)
	 retVal = LoadIntMapFromMemory(hFileData, x, y, w, h, ix, iy, iw, ih);

	 // Free the memory and return
	 GlobalFree(hFileData);
	 return retVal;
  }

  // File length was zero :-(
  fclose(hFile);
  return FALSE;
}

BYTE IntMap::NearestColor(COLORREF target, BYTE palEntry, LPRGBTRIPLE lpPalette, LPBYTE lpMergeDecoder, BYTE nEntries, BOOL modifyDecoder) {
  // Match the nearest color in the palette
  WORD leastDiff = 768; // 256 * 3, a value never achieved
  WORD leastDiffEntry;
  WORD curDiff;
  BYTE i; // largest palette size is 256, so BYTE is used

  // Check every palette entry
  for (i = 0; i < nEntries; i++) {
	 curDiff = (GetRValue(target) > lpPalette[i]->rgbtRed)? GetRValue(target) - lpPalette[i]->rgbtReg: lpPalette[i]->rgbtRed - GetRValue(target);
	 curDiff += (GetGValue(target) > lpPalette[i]->rgbtGreen)? GetGValue(target) - lpPalette[i]->rgbtReg: lpPalette[i]->rgbtGreen - GetGValue(target);
	 curDiff += (GetBValue(target) > lpPalette[i]->rgbtBlue)? GetBValue(target) - lpPalette[i]->rgbtReg: lpPalette[i]->rgbtBlue - GetBValue(target);
	 if (curDiff < leastDiff) {
		leastDiff = curDiff;
		leastDiffEntry = 0;
	 }
  }

  // Modify the mergeDecoder, if we need to
  if (modifyDecoder)
	 lpMergeDecoder[palEntry] = leastDiffEntry;

  // Return the nearest color :-)!
  return leastDiffEntry;
}

BOOL IntMap::LoadIntMapFromMemory(HGLOBAL hFileData, UINT x, UINT y, UINT w, UINT h, UINT ix, UINT iy, WORD wFlags) {
  LPWORD lpFileHeader;

  // Check the file type
  lpFileHeader = (LPWORD) GlobalLockEx(hFileData, TRUE, 0, sizeof(WORD));
  if (*lpFileHeader == 19778) {
	 // This is a bitmap :-)!
	 LPBITMAPFILEHEADER lpBitmapFileHeader;
	 LPBITMAPINFO lpBitmapInfo;
	 LPRGBQUAD lpRGB;
	 LPRGBTRIPLE lpPalette;
	 DWORD memPos;
	 DWORD nColors;
	 DWORD i, curX, curY;
	 LPBYTE lpFileData;
	 BYTE bitCount;
	 UINT bmpWidth, bmpHeight;
	 BYTE mergeDecoder[256];
	 UINT widthInBytes;

	 // Set the mergeDecoder to the default values
	 for (i = 0; i < 256; i++)
		mergeDecoder[i] = i;

	 // Get the headers of the bitmap
	 GlobalUnlock(hFileData);
	 lpBitmapFileHeader = (LPBITMAPFILEHEADER) GlobalLockEx(hFileData, TRUE, 0, sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFO));
	 lpBitmapInfo = (LPBITMAPINFO) ((DWORD) lpBitmapFileHeader + sizeof(BITMAPFILEHEADER));
	 memPos = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFO);

	 // Get the bitmap width and height
	 bmpWidth = (UINT) lpBitmapInfo->bmiHeader.biWidth;
	 bmpHeight = (UINT) lpBitmapInfo->bmiHeader.biHeight;

	 // Set the defaults if the user specified zero for the parameters
	 if (bmpWidth - 1 < x)    						x = 0;
	 if (bmpHeight - 1 < y)	 						y = 0;
	 if ((w == 0) || (bmpWidth < w + x))  		w = bmpWidth - x;
	 if ((h == 0) || (bmpHeight < h + y))		h = bmpHeight - y;

	 // If this file is larger than the current allocation (ie, if
	 // there is a current allocation), then we'll have to resize it
	 if ((bmpWidth > width) || (bmpHeight > height)) {
		if (hIntMapData) {
		  HGLOBAL hNewBackgroundData;
		  if (criticalSectionOn) {
			 criticalSectionOn = FALSE;
			 GlobalArrayEnd(hIntMapData);
			 GlobalArrayEnd(hBackgroundData);
		  }
		  GlobalFree(hIntMapData);

		  // Resize the background, not caring about the data inside
		  hNewBackgroundData = GlobalReAlloc(hBackgroundData, bmpWidth * bmpHeight, GMEM_MOVEABLE | GMEM_ZEROINIT);
		  if (!hNewBackgroundData) {
			 GlobalUnlock(hFileData);
			 return FALSE;
		  }
		  hBackgroundData = hNewBackgroundData;
		} else {
		  // Allocate new background memory
		  hBackgroundData = GlobalAlloc(GHND, bmpWidth * bmpHeight);
		  if (!hBackgroundData) {
			 GlobalUnlock(hFileData);
			 return FALSE;
		  }
		}

		// Allocate new memory
		hIntMapData = GlobalAlloc(GHND, bmpWidth * bmpHeight);
		if (!hIntMapData) {
		  GlobalUnlock(hFileData);
		  return FALSE;
		}

		// Adjust width and height variables
		width = bmpWidth;
		height = bmpHeight;
	 }

	 // Check to see if the location parameters are valid
	 if ((width - 1 < ix) || (width < w + ix))	ix = 0;
	 if ((height - 1 < iy) || (height < h + iy)) iy = 0;

	 // Get the colors
	 bitCount = lpBitmapInfo->bmiHeader.biBitCount;
	 nColors = 1 << bitCount;

	 // Check to see if we support this type of bitmap
	 if ((lpBitmapInfo->biCompression != BI_RGB) || (nColors == 0) || (nColors > getpalettesize()) || (bitCount >= 32)){
		// Unsupported bitmap type (unsupported compression, no colors,
		// 32bpp allocation, or too many colors)
		GlobalUnlock(hFileData);
		return FALSE;
	 }

	 // Lock the entire palette
	 lpPalette = GlobalLock(hPalette);
	 if (!lpPalette) {
		GlobalUnlock(hFileData);
		return FALSE;
	 }

	 // Load the palette if we have one
	 if ((DWORD) lpBitmapInfo->bmiColors != NULL) {
		// Get the colors
		GlobalUnlock(hFileData);
		lpRGB = (LPRGBQUAD) GlobalLockEx(hFileData, TRUE, memPos - sizeof(RGBQUAD), sizeof(RGBQUAD) * nColors);
		memPos += sizeof(RGBQUAD) * nColors;

		if (wFlags & INTMAP_PRESERVESYSTEMPALETTE) {
		  if (wFlags & INTMAP_PRESERVEENTIREPALETTE) {
			 // Transfer the colors from the file into the palette, keeping
			 // all the entries intact by simply finding the nearest color
			 for (i = 0; i < nColors; i++) {
				NearestColor(RGB(*lpRGB.rgbRed, *lpRGB.rgbGreen, *lpRGB.rgbBlue), i, lpPalette, mergeDecoder, nColors, TRUE);
				lpRGB = (LPRGBQUAD) ((DWORD) lpRGB + sizeof(RGBQUAD));
			 }
		  } else {
			 // Transfer the colors from the file into the palette, keeping
			 // the first entries intact and finding the nearest color
			 // for the rest
			 lpPalette = (LPRGBQUAD) ((DWORD) lpPalette + 16 * sizeof(RGBTRIPLE));
			 for (i = 0; i < nColors - 16; i++) {
				*lpPalette.rgbtRed = *lpRGB.rgbRed;
				*lpPalette.rgbtGreen = *lpRGB.rgbGreen;
				*lpPalette.rgbtBlue = *lpRGB.rgbBlue;
				mergeDecoder[i] = i + 16;
				lpPalette = (LPRGBTRIPLE) ((DWORD) lpPalette + sizeof(RGBTRIPLE));
				lpRGB = (LPRGBQUAD) ((DWORD) lpRGB + sizeof(RGBQUAD));
			 }

			 // Match the nearest colors here
			 lpPalette = (LPRGBTRIPLE) ((DWORD) lpPalette - nColors * sizeof(RGBTRIPLE));
			 for (i = nColors - 16; i < nColors; i++) {
				NearestColor(RGB(*lpRGB.rgbRed, *lpRGB.rgbGreen, *lpRGB.rgbBlue), i, lpPalette, mergeDecoder, nColors, TRUE);
				lpRGB = (LPRGBQUAD) ((DWORD) lpRGB + sizeof(RGBQUAD));
			 }
		  }
		} else {
		  // Transfer the colors from the file into the palette, verbatim
		  for (i = 0; i < nColors; i++) {
			 *lpPalette.rgbtRed = *lpRGB.rgbRed;
			 *lpPalette.rgbtGreen = *lpRGB.rgbGreen;
			 *lpPalette.rgbtBlue = *lpRGB.rgbBlue;
			 lpPalette = (LPRGBTRIPLE) ((DWORD) lpPalette + sizeof(RGBTRIPLE));
			 lpRGB = (LPRGBQUAD) ((DWORD) lpRGB + sizeof(RGBQUAD));
		  }
		}

		// Unlock bitmap file palette
		GlobalUnlock(hFileData);
	 } else
		GlobalUnlock(hFileData);

	 // Start the global arrays for the IntMapData
	 widthInBytes = (bmpWidth * bitCount) / 8;
	 if ((bmpWidth * bitCount) % 8 > 0)
		++widthInBytes;
	 widthInBytes = (widthInBytes % 4 > 0)? widthInBytes + (4 - widthInBytes % 4): widthInBytes;
	 if (!GlobalArrayStart(hFileData, widthInBytes, bmpHeight, 1)) {
		GlobalUnlock(hPalette);
		GlobalFree(hIntMapData);
		hIntMapData = NULL;
		return FALSE;
	 }
	 if (!criticalSectionOn) {
		if (!GlobalArrayStart(hIntMapData, width, height, (maxConventionalSize / 2) / width)) {
		  GlobalArrayEnd(hFileData);
		  GlobalFree(hIntMapData);
		  GlobalUnlock(hPalette);
		  hIntMapData = NULL;
		  return FALSE;
		}
	 }

	 // Get the bitmap bits
	 for (curY = h + y - 1; curY >= y; curY--) {
		lpFileData = (LPBYTE) GlobalArrayElement(hFileData, curY);
		lpIntMapData = (LPBYTE) GlobalArrayElement(hIntMapData, curY - y + iy);
		switch (bitCount) {
		  case 1: {
			 // 1-bit bitmap
			 for (curX = x; curX < x + w; curX++)
				lpIntMapData[curX - x + ix] = mergeDecoder[lpFileData[curX / 8] & (1 << (curX % 8))];
			 break;
		  }
		  case 2: {
			 // 2-bit bitmap (enables reading of Windows CE bitmaps)
			 for (curX = x; curX < x + w; curX++)
				lpIntMapData[curX - x + ix] = mergeDecoder[lpFileData[curX / 4] & (3 << (2 * (curX % 4)))];
			 break;
		  }
		  case 8: {
			 // 8-bit bitmap
			 for (curX = x; curX < x + w; curX++)
				lpIntMapData[curX - x + ix] = mergeDecoder[lpFileData[curX]];
			 break;
		  }
		  case 24: {
			 // 24-bit bitmap
			 for (curX = x; curX < x + w; curX++)
				lpIntMapData[curX - x + ix] = NearestColor(RGB(lpFileData[curX * 3], lpFileData[curX * 3 + 1], lpFileData[curX * 3 + 2]), NULL, lpPalette, NULL, nColors, FALSE);
			 break;
		  }
		  default: {
			 // Unsupported type; return FALSE after cleaning up
			 GlobalArrayEnd(lpFileData);
			 GlobalArrayEnd(lpIntMapData);
			 GlobalUnlock(hPalette);
			 return FALSE;
		  }
		}
	 }

	 // Unlock the IntMap palette
	 GlobalUnlock(hPalette);

	 // Unlock the bitmap data
	 GlobalUnlock(hFileData);

	 // Process the critical section stuff
	 if (!criticalSection) {
		criticalSectionOn = FALSE;
		GlobalArrayEnd(hIntMapData);
	 }
	 if (criticalSection && !criticalSectionOn) {
		criticalSectionOn = TRUE;
		GlobalArrayStart(hBackgroundData, width, height, (maxConventionalSize / 2) / width);
	 }

	 // We did it!  We loaded a bitmap into the IntMap!
	 return TRUE;
  }
  if (((lpFileHeader >> 8) >= '0') && ((lpFileHeader >> 8) <= '9'))
  {
	 // This is an IntMap1 file :-)!
	 HGLOBAL hNewBackgroundData;
	 UINT curX, curY;

	 // Get rid of any previous allocation
/*	 if (hIntMapData) {
		if (criticalSectionOn) {
		  criticalSectionOn = FALSE;
		  GlobalArrayEnd(hIntMapData);
		  GlobalArrayEnd(hBackgroundData);
		}
		GlobalFree(hIntMapData);
		hIntMapData = NULL;
	 }*/

	 // Open IntMap file and read the number of elements
	 lpFileData = GlobalLock(hFileData);
	 sscanf(lpFileData, "%u %u", &impWidth, &impHeight);
	 impWidth++;
	 impHeight++;


  // Resize the background, not caring about the data inside
/*  hNewBackgroundData = GlobalReAlloc(hBackgroundData, width * height, GMEM_MOVEABLE | GMEM_ZEROINIT);
  if (!hNewBackgroundData) {
	 fclose(hFile);
	 return FALSE;
  }
  hBackgroundData = hNewBackgroundData;*/

	 // Allocate new memory
	 hIntMapData = GlobalAlloc(GHND, width * height);
	 if (!hIntMapData) {
		GlobalUnlock(hFileData);
		return FALSE;
	 }
	 if (!GlobalArrayStart(hIntMapData, width, height, (maxConventionalSize / 2) / width)) {
		GlobalFree(hIntMapData);
		GlobalUnlock(hFileData);
		hIntMapData = NULL;
		return FALSE;
	 }

	 // Read the IntMap into memory (change the background color, if
	 // detected to the current background color index)
	 for (y = 0; y < impHeight; y++) {
		for (x = 0; x < width; x++) {
		  scanf("%hu", &lpIntMapData[x]);
		  if (lpIntMapData[x] == 20) // 20 is the IntMap1 background color
			 lpIntMapData[x] = backgroundColor;
		}
    }

  // Close the file
  fclose(hFile);

  // If the programmer wanted a critical section, then don't end the global array
  if (criticalSection) {
	 criticalSectionOn = TRUE;
	 GlobalArrayStart(hBackgroundData, width, height, (maxConventionalSize / 2) / width);
  } else {
	 criticalSectionOn = FALSE;
	 GlobalArrayEnd(hIntMapData);
  }
  }




	 lpRGB = (LPRGBQUAD) GlobalLockEx(hFileData, TRUE, memPos - sizeof(RGBQUAD), sizeof(RGBQUAD) * nColors);
	 memPos += sizeof(RGBQUAD) * nColors;

		pDstPal->palPalEntry[Count].peRed   = pRgb[Count].rgbRed;
		lpDstPal->palPalEntry[Count].peGreen = pRgb[Count].rgbGreen;
		lpDstPal->palPalEntry[Count].peBlue  = pRgb[Count].rgbBlue;
		lpDstPal->palPalEntry[Count].peFlags = 0;
	 }
	hNewPalette = CreatePalette(lpDstPal);
	GlobalFree( (HANDLE) GlobalHandle( HIWORD( lpDstPal) ) );
	hOldPalette = SelectPalette(hDC,hNewPalette,FALSE);
	RealizePalette(hDC);
	BitmapHandle = CreateDIBitmap(hDC, (LPBITMAPINFOHEADER)pBits, CBM_INIT, ((LPSTR) pBits + 40) + (nColors * sizeof(RGBQUAD)), pBits, DIB_RGB_COLORS);
	SelectPalette(hDC,hOldPalette,FALSE);
	FreeResource(hBitmap);
	ReleaseDC(hWnd,hDC);
	DeleteObject(hOldPalette);
	}

      return BitmapHandle;



  scanf("%u %u", &width, &height);
  width++;
  height++;

  // Resize the background, not caring about the data inside
  hNewBackgroundData = GlobalReAlloc(hBackgroundData, width * height, GMEM_MOVEABLE | GMEM_ZEROINIT);
  if (!hNewBackgroundData) {
	 fclose(hFile);
	 return FALSE;
  }
  hBackgroundData = hNewBackgroundData;

  // Allocate new memory
  hIntMapData = GlobalAlloc(GHND, width * height);
  if (!hIntMapData) {
	 fclose(hFile);
	 return FALSE;
  }
  if (!GlobalArrayStart(hIntMapData, width, height, (maxConventionalSize / 2) / width)) {
	 GlobalFree(hIntMapData);
	 fclose(hFile);
	 hIntMapData = NULL;
	 return FALSE;
  }

  // Read the IntMap into memory (after making the background color 20)
  backgroundColor = 20;
  for (y = 0; y < height; y++) {
	 lpIntMapData = (LPBYTE) GlobalArrayElement(hIntMapData, y);
	 for (x = 0; x < width; x++)
		scanf("%hu", &lpIntMapData[x]);
  }

  // Close the file
  fclose(hFile);

  // If the programmer wanted a critical section, then don't end the global array
  if (criticalSection) {
	 criticalSectionOn = TRUE;
	 GlobalArrayStart(hBackgroundData, width, height, (maxConventionalSize / 2) / width);
  } else {
	 criticalSectionOn = FALSE;
	 GlobalArrayEnd(hIntMapData);
  }
}

BOOL IntMap::PasteIntMap(UINT x, UINT y, UINT w, UINT h) {
  UINT curX, curY;
  UINT IntMapColor, ScreenMapColor, ScreenColor;

  // If we are not in a critical section, then start the global array
  if (!criticalSectionOn) {
	 if (!GlobalArrayStart(hIntMapData, width, height, (maxConventionalSize / 2) / width))
		return FALSE;
	 if (!GlobalArrayStart(hBackgroundData, width, height, (maxConventionalSize / 2) / width)) {
		GlobalArrayEnd(hIntMapData);
		return FALSE;
	 }
  }

  // Loop through all the pixels in the replacement IntMap
  for (curY = 0; curY < h; curY++) {
	 // Get the current row into memory
	 lpIntMapData = GlobalArrayElement(hIntMapData, curY);
	 lpBackgroundData = GlobalArrayElement(hBackgroundData, curY);

	 // Go through the columns
	 for (curX = 0; curX < w; curX++) {
		// Initialize color values
		IntMapColor := *((LPBYTE) ((DWORD) lpIntMapData + curX));
		ScreenMapColor := *((LPBYTE) ((DWORD) lpBackgroundData + curX));
		ScreenColor := getpixel(x + curX, y + curY);

		// If the screen color is not what is needs to be, change it
		if ((ScreenColor != IntMapColor) ||	(ScreenColor != ScreenMapColor))
		  if (IntMapColor = clear)
			 putpixel(x + curX, y + curY, ScreenMapColor);
		  else
			 putpixel(x + curX, y + curY, IntMapColor);
	 }
  }

  // If we are supposed to end the critical section, then do so; if we are
  // supposed to begin it, then do so
  if (criticalSection && !criticalSectionOn)
	 criticalSectionOn = TRUE;
  if (!criticalSection) {
	 criticalSectionOn = FALSE;
	 GlobalArrayEnd(hIntMapData);
	 GlobalArrayEnd(hBackgroundData);
  }
  return TRUE;
}

BOOL IntMap::EraseRegion(UINT x, UINT y, UINT w, UINT h) {
  UINT curX, curY;

  // If we are not in a critical section, then start the global array
  if (!criticalSectionOn)
	 if (!GlobalArrayStart(hBackgroundData, width, height, (maxConventionalSize / 2) / width))
		return FALSE;

  // Load the ScreenIntMap into the region given
  for (curY = 0; curY < h; curY++) {
	 // Get the current row into memory
	 lpBackgroundData = GlobalArrayElement(hBackgroundData, curY);

	 // Write the background data to the screen
	 for (curX = 0; curX < w; curX++)
		putpixel(curX + x, curY + y, *((LPBYTE) ((DWORD) lpBackgroundData + curX)));
  }

  // If the programmer wanted a critical section, then don't end the global array
  if (criticalSection && !criticalSectionOn) {
	 criticalSectionOn = TRUE;
	 GlobalArrayStart(hIntMapData, width, height, (maxConventionalSize / 2) / width);
  }
  if (!criticalSection) {
	 criticalSectionOn = FALSE;
	 GlobalArrayEnd(hBackgroundData);
  }
  return TRUE;
}

BOOL IntMap::ClippedEraseRegion(UINT x, UINT y, UINT w, UINT h, UINT cx, UINT cy, UINT cw, UINT ch) {
  UINT curX, curY;

  // If we are not in a critical section, then start the global array
  if (!criticalSectionOn)
	 if (!GlobalArrayStart(hBackgroundData, width, height, (maxConventionalSize / 2) / width))
		return FALSE;

  // Load the ScreenIntMap into the region given
  for (curY = 0; curY < h; curY++) {
	 // Get the current row into memory
	 lpBackgroundData = GlobalArrayElement(hBackgroundData, curY);

	 // Write the background data to the screen
	 for (curX = 0; curX < w; curX++)
		if (!((curX + x >= cx) && (curX + x <= cx + cw) && (curY + y >= cy) && (curY + y <= cy + ch)))
		  putpixel(curX + x, curY + y, *((LPBYTE) ((DWORD) lpBackgroundData + curX)));
  }

  // If the programmer wanted a critical section, then don't end the global array
  if (criticalSection && !criticalSectionOn) {
	 criticalSectionOn = TRUE;
	 GlobalArrayStart(hIntMapData, width, height, (maxConventionalSize / 2) / width);
  }
  if (!criticalSection) {
	 criticalSectionOn = FALSE;
	 GlobalArrayEnd(hBackgroundData);
  }
  return TRUE;
}

BOOL IntMap::GetScreenIntMap(UINT sx, UINT sy, UINT ix, UINT iy, UINT iw, UINT ih) {
  UINT curX, curY;

  // If we are not in a critical section, then start the global array
  if (!criticalSectionOn)
	 if (!GlobalArrayStart(hBackgroundData, width, height, (maxConventionalSize / 2) / width))
		return FALSE;

  // Load the ScreenIntMap from the region given
  for (curY = iy; curY < iy + ih; curY++) {
	 // Get the current row into memory
	 lpBackgroundData = GlobalArrayElement(hBackgroundData, curY);

	 // Get the background data from the screen
	 for (curX = ix; curX < ix + iw; curX++)
		*((LPBYTE) ((DWORD) lpBackgroundData + curX)) = getpixel(curX + sx, curY + sy);
  }

  // If the programmer wanted a critical section, then don't end the global array
  if (criticalSection && !criticalSectionOn) {
	 criticalSectionOn = TRUE;
	 GlobalArrayStart(hIntMapData, width, height, (maxConventionalSize / 2) / width);
  }
  if (!criticalSection) {
	 criticalSectionOn = FALSE;
	 GlobalArrayEnd(hBackgroundData);
  }
  return TRUE;
}

BOOL IntMap::GetIntMapFromScreen(UINT sx, UINT sy, UINT ix, UINT iy, UINT iw, UINT ih) {
  UINT curX, curY;
  BYTE color;

  // If we are not in a critical section, then start the global array
  if (!criticalSectionOn)
	 if (!GlobalArrayStart(hIntMapData, width, height, (maxConventionalSize / 2) / width))
		return FALSE;

  // Load the IntMap from the region given
  for (curY = iy; curY < iy + ih; curY++) {
	 // Get the current row into memory
	 lpIntMapData = GlobalArrayElement(hIntMapData, curY);

	 // Get the background data from the screen
	 for (curX = ix; curX < ix + iw; curX++)
		*((LPBYTE) ((DWORD) lpIntMapData + curX)) = getpixel(curX + sx, curY + sy);
  }

  // If the programmer wanted a critical section, then don't end the global array
  if (criticalSection && !criticalSectionOn) {
	 criticalSectionOn = TRUE;
	 GlobalArrayStart(hBackgroundData, width, height, (maxConventionalSize / 2) / width);
  }
  if (!criticalSection) {
	 criticalSectionOn = FALSE;
	 GlobalArrayEnd(hIntMapData);
  }
  return TRUE;
}

static BOOL IntMap::CopyIntMap(UINT x1, UINT y1, UINT w, UINT h, UINT x2, UINT y2, IntMap NewIntMap, IntMap OldIntMap) {
  UINT curX, curY;

  // Copy the data
  // If they are not in a critical sections, then start the global arrays
  if (!NewIntMap->criticalSectionOn)
	 if (!GlobalArrayStart(NewIntMaphBackgroundData, width, height, (maxConventionalSize / 2) / width))
		return FALSE;

  // Load the ScreenIntMap into the region given
  for (curY = iy; curY < ih; curY++) {
	 // Get the current row into memory
	 lpIntMapData = GlobalArrayElement(hIntMapData, curY);
	 lpBackgroundData = GlobalArrayElement(hBackgroundData, curY);

	 // Get the background data from the screen
	 for (curX = ix; curX < iw; curX++)
		if (!((curX + x >= cx) && (curX + x <= cx + cw) && (curY + y >= cy) && (curY + y <= cy + ch)))
		  *((LPBYTE) ((DWORD) lpBackgroundData + curX)) = getpixel(curX + sx, curY + sy);
  }

  // If the programmer wanted a critical section, then don't end the global array
  if (criticalSection && !criticalSectionOn) {
	 criticalSectionOn = TRUE;
	 GlobalArrayStart(hIntMapData, width, height, (maxConventionalSize / 2) / width);
  }
  if (!criticalSection) {
	 criticalSectionOn = FALSE;
	 GlobalArrayEnd(hBackgroundData);
  }
  return TRUE;
}
		VAR
			curx, cury : integer;
		BEGIN
			FOR curx := 0 TO w DO
				FOR cury := 0 TO h DO
					NewIntMap[curx + x1, cury + y1] := OldIntMap[curx + x2, cury + y2];
		END;

	PROCEDURE IntersectRect(x1, y1, w1, h1, x2, y2, w2, h2 : integer; VAR xi, yi, wi, hi : integer; VAR Intersect : boolean);
		VAR
			curx, cury : integer;
			IntersectIntMap : PACKED ARRAY [0..IntMapWidth, 0..IntMapHeight] OF boolean;
		BEGIN
			Intersect := false;
			FOR cury := y1 TO y1 + h1 DO
            FOR curx := x1 TO x1 + w1 DO
               IF (curx >= x2) AND (curx <= x2 + w2) AND (cury >= y2) AND (cury <= y2 + h2) THEN
                  BEGIN
                     IntersectIntMap[curx - x1, cury - y1] := true;
                     Intersect := true;
						END
               ELSE
                  IntersectIntMap[curx - x1, cury - y1] := false;

         IF Intersect THEN
            BEGIN
               Intersect := false;
               curx := 0;
               cury := 0;
               REPEAT
                  IF IntersectIntMap[curx, cury] THEN
                     Intersect := true
                  ELSE
                     IF (curx + 1) > w1 THEN
                        BEGIN
                           curx := 0;
									cury := cury + 1;
                           IF cury > h1 THEN
										BEGIN
                                 writeln('IntersectRect Internal Failure');
                                 Intersect := true;
                              END;
                        END
                     ELSE
                        curx := curx + 1;
               UNTIL Intersect;

               IF cury <= h1 THEN
                  BEGIN
                     xi := curx;
                     yi := cury;

							curx := curx + 1;
                     Intersect := false;
                     REPEAT
                        IF curx > w1 THEN
									BEGIN
                              curx := curx - 1;
                              Intersect := true;
                           END
                        ELSE
                           IF IntersectIntMap[curx, cury] = false THEN
                              BEGIN
                                 curx := curx - 1;
                                 Intersect := true;
                              END
                           ELSE
                              curx := curx + 1;
							UNTIL Intersect;

                     wi := curx - xi;

                     cury := cury + 1;
                     curx := xi;
							Intersect := false;
                     REPEAT
                        IF cury > h1 THEN
                           BEGIN
                              cury := cury - 1;
                              Intersect := true;
                           END
                        ELSE
                           IF IntersectIntMap[curx, cury] = false THEN
                              BEGIN
											cury := cury - 1;
                                 Intersect := true;
                              END
                           ELSE
                              cury := cury + 1;
                     UNTIL Intersect;

                     hi := cury - yi;
							xi := xi + x1;
                     yi := yi + y1;
                  END;
            END;
      END;


   PROCEDURE MoveIntMap(x1, y1, w1, h1, x2, y2, w2, h2 : integer; IntMap : IntMapField; VAR OldIntMap : IntMapField);
		VAR
         curx, cury : integer;
         Intersect : boolean;
         xi, yi, wi, hi : integer;
         NewIntMap : IntMapField;
      BEGIN
         IntersectRect(x2, y2, w2, h2, x1, y1, w1, h1, xi, yi, wi, hi, Intersect);
         IF Intersect THEN
            BEGIN
               GetIntMapFromScreen(x2, y2, 0, 0, w2, h2, NewIntMap);
					CopyIntMap(xi - x2, yi - y2, wi, hi, xi - x1, yi - y1, NewIntMap, OldIntMap);
               ClippedEraseRegion(x1, y1, w1, h1, xi, yi, wi, hi, OldIntMap);
               PasteIntMap(x2, y2, w2, h2, IntMap, NewIntMap);
               OldIntMap := NewIntMap;
            END
         ELSE
				BEGIN
               GetIntMapFromScreen(x2, y2, 0, 0, w2, h2, NewIntMap);
               EraseRegion(x1, y1, w1, h1, OldIntMap);
               PasteIntMap(x2, y2, w2, h2, IntMap, NewIntMap);
               OldIntMap := NewIntMap;
            END;
		END;

	PROCEDURE RotateIntMap(x, y : integer; VAR w, h : integer; VAR IntMap : IntMapField; RotateLeft : boolean);
		VAR
			NewIntMap : IntMapField;
			curx, cury : integer;
			newx, newy : integer;
			Temp : integer;
		BEGIN
			NewIntMap := IntMap;
			newy := y;

			IF RotateLeft THEN
				FOR curx := x + w DOWNTO x DO
					BEGIN
						newx := x;
						FOR cury := y TO y + h DO
							BEGIN
                        NewIntMap[newx, newy] := IntMap[curx, cury];
                        newx := newx + 1;
                     END;
						newy := newy + 1;
               END
         ELSE
				FOR curx := x TO x + w DO
               BEGIN
						newx := x;
                  FOR cury := y + h DOWNTO y DO
                     BEGIN
                        NewIntMap[newx, newy] := IntMap[curx, cury];
                        newx := newx + 1;
                     END;
                  newy := newy + 1;
               END;

         Temp := w;
         w := h;
         h := Temp;

         IntMap := NewIntMap;
      END;

END.