#include <dos.h>
#include <mem.h>
#include <stdio.h>
#include <stdlib.h>
#include <alloc.h>
#include <stdarg.h>
#include "winmem.h"


/* Global Memory Flags */
#define GMEM_FIXED          0x0000
#define GMEM_XMS            0x0001 // Object is in XMS
#define GMEM_MOVEABLE       0x0002
#define GMEM_EMS				 0x0004 // Object is in EMS
#define GMEM_CONVENTIONAL   0x0008 // Object is in conventional;
#define GMEM_NOCOMPACT      0x0010   // if GMEM_CONVENTIONAL is set
#define GMEM_NODISCARD      0x0020   // and GMEM_?MS is set, the data
#define GMEM_ZEROINIT       0x0040   // is temporarily in conventional;
#define GMEM_MODIFY         0x0080   // otherwise, memory is set in
#define GMEM_DISCARDABLE    0x0100   // conventional with malloc
#define GMEM_NOT_BANKED     0x1000
#define GMEM_SHARE          0x2000
#define GMEM_DDESHARE       0x2000
#define GMEM_NOTIFY         0x4000
#define GMEM_LOWER          GMEM_NOT_BANKED
#define GMEM_VALID_FLAGS    0x7F7F
#define GMEM_INVALID_HANDLE 0x8000

#define GHND                (GMEM_MOVEABLE | GMEM_ZEROINIT)
#define GPTR                (GMEM_FIXED | GMEM_ZEROINIT)

#define GlobalLRUNewest( h )    (HANDLE)(h)
#define GlobalLRUOldest( h )    (HANDLE)(h)
#define GlobalDiscard( h )      GlobalReAlloc( (h), 0, GMEM_MOVEABLE )

/* Flags returned by GlobalFlags (in addition to GMEM_DISCARDABLE) */
#define GMEM_DISCARDED      0x4000
#define GMEM_LOCKCOUNT      0x00FF

typedef struct _MEMORYSTATUS {
	 DWORD dwLength;
	 DWORD dwMemoryLoad;
	 DWORD dwTotalPhys;
	 DWORD dwAvailPhys;
	 DWORD dwTotalPageFile;
	 DWORD dwAvailPageFile;
	 DWORD dwTotalVirtual;
	 DWORD dwAvailVirtual;
	 DWORD fMemTypes; // includes the types of memory in the system
} MEMORYSTATUS, *LPMEMORYSTATUS;

typedef struct _XMS_MOVEPARAMS {
  DWORD blockLength;  // size of block, in bytes
  WORD sourceHandle;
  DWORD sourcePtr;
  WORD destHandle;
  DWORD destPtr;
} XMS_MOVEPARAMS;
XMS_MOVEPARAMS xmsMoveParams;

typedef struct _MEMOBJ_LIST {
  HGLOBAL hglbMem;
  LPVOID lockedMemory;
  DWORD locked;
  DWORD fuFlags;
  DWORD cbBytes;

  // Used for internal locking and unlocking variables
  DWORD param1;
  DWORD param2;
  WORD param3;

  // Used for the GlobalArray feature
  DWORD dwElementSize;
  DWORD nElements;
  DWORD curElement;
  BOOL arrayLocked;
  DWORD nCachedElements;
} MEMOBJ_LIST, *LPMEMOBJ_LIST;

LPMEMOBJ_LIST lpMemObjList;
WORD nHandles;
WORD nA20Locks;

void far (*XMSFunc) (void);

// Forward references for functions

extern HGLOBAL GlobalAlloc(DWORD fuFlags, DWORD cbBytes);
extern HGLOBAL GlobalFree(HGLOBAL hglbMem);
extern HGLOBAL GlobalReAlloc(HGLOBAL hglbMem, DWORD cbBytes, DWORD fuFlags);
extern LPVOID GlobalLockEx(HGLOBAL hglbMem, BOOL conventional, DWORD start, DWORD len);  // DWORD start (0) - specifies offset of the beginning of the block to place in conventional if conventional is TRUE
																				 // DWORD len (0) - specifies length of block to copy to conventional if conventional is true; 0 copies entire block from start to the end of the memory allocation
BOOL ForceUnlock(HGLOBAL hglbMem);
extern DWORD GlobalFlags(HGLOBAL hglbMem);
extern HGLOBAL GlobalHandle(LPCVOID lpvMem);
extern BOOL GlobalUnlock(HGLOBAL hglbMem);
extern VOID GlobalMemoryStatus(LPMEMORYSTATUS lpmstMemStat);
extern DWORD GlobalSize(HGLOBAL hglbMem);
extern BOOL GlobalStart(DWORD memObjListFlags); // Start a KittyRAM session
extern DWORD GlobalEnd(VOID);  // End a KittyRAM session
extern BOOL GlobalReady(VOID);
extern BOOL GlobalArrayStart(HGLOBAL hglbMem, DWORD dwElementSize, DWORD nElements, DWORD nCachedElements);
extern LPVOID GlobalArrayFirst(HGLOBAL hglbMem);
extern LPVOID GlobalArrayNext(HGLOBAL hglbMem);
extern LPVOID GlobalArrayElement(HGLOBAL hglbMem, DWORD dwElement);
extern BOOL GlobalArrayEnd(HGLOBAL hglbMem);
extern VOID ZeroMemory(LPVOID Destination, DWORD Length);
extern VOID FillMemory(LPVOID Destination, DWORD Length, BYTE Fill);

BOOL XMSExists; // Does this system have XMS?
BOOL EMSExists; // Does this system have EMS?
BOOL init = FALSE; // Have we been initialized?

// Beginning of library

/*typedef struct XMM_Move {
	 DWORD	Length;			//  Number of bytes to move
	 WORD	SourceHandle;		//  Source handle
	 DWORD	SourceOffset;		//  Source offset into block memory
	 WORD	DestHandle;		//  Destination handle
	 DWORD	DestOffset;		//  Destination offset into block memory
} XMM_Move;
XMM_Move XMM_MoveStruct;

WORD XMM_MoveMemory(DWORD length, WORD srchnd, DWORD srcoff, WORD dsthnd, DWORD dstoff)
{
  WORD MoveStructOff = FP_OFF(&XMM_MoveStruct);
  WORD RetVal;

  XMM_MoveStruct.Length = length;
  XMM_MoveStruct.SourceHandle = srchnd;
  XMM_MoveStruct.SourceOffset = srcoff;
  XMM_MoveStruct.DestHandle = dsthnd;
  XMM_MoveStruct.DestOffset = dstoff;

  asm {
	 mov	si, MoveStructOff
	 mov	ah, 0x0B
	 call	[XMSFunc]
	 mov	RetVal, ax
  }

  return(RetVal);
}*/

LPMEMOBJ_LIST GetMemObjList(HGLOBAL hglbMem, LPVOID lockedMemory, LPDWORD lpByteOffset, LPWORD lpIndex) {
  // Find the memory object in the list and return the list data
  WORD i = 0;

  // Search the list
  while (i < nHandles) {
	 if ((lpMemObjList[i].hglbMem == hglbMem) || (lpMemObjList[i].lockedMemory == lockedMemory)) {
		// We found it!  Return this object description!
		if (lpByteOffset)
		  *lpByteOffset = &lpMemObjList[i] - lpMemObjList;
		if (lpIndex)
		  *lpIndex = i;
		return &lpMemObjList[i];
	 }
	 i++;
  }

  // Not found; return NULL
  return NULL;
}

BOOL AddMemory(HGLOBAL hglbMem, DWORD cbBytes, DWORD fuFlags) {
  // Adds this memory object to our list of memory objects

  // Lock the memory
  LPVOID newList;

  // Increase the memory size of the memory object list
  if (!lpMemObjList)
	 newList = (LPVOID) farmalloc(sizeof(MEMOBJ_LIST));
  else
	 newList = (LPVOID) farrealloc(lpMemObjList, (nHandles + 1) * sizeof(MEMOBJ_LIST));

  // See if the allocation suceeded
  if (!newList)
	 return FALSE;
  lpMemObjList = (LPMEMOBJ_LIST) newList;

  // Add our new member to the list
  lpMemObjList[nHandles].hglbMem = hglbMem;
  lpMemObjList[nHandles].lockedMemory = 0;
  lpMemObjList[nHandles].locked = FALSE;
  lpMemObjList[nHandles].fuFlags = fuFlags;
  lpMemObjList[nHandles].cbBytes = cbBytes;
  lpMemObjList[nHandles].param1 = 0;
  lpMemObjList[nHandles].param2 = 0;
  lpMemObjList[nHandles].param3 = 0;
  lpMemObjList[nHandles].dwElementSize = 0;
  lpMemObjList[nHandles].nElements = 0;
  lpMemObjList[nHandles].curElement = 0;
  lpMemObjList[nHandles].arrayLocked = 0;
  lpMemObjList[nHandles].nCachedElements = 0;

  // Increment the list size
  nHandles++;

  // Lock the new allocation and return the pointer
  return TRUE;
}

extern HGLOBAL GlobalAlloc(DWORD fuFlags, DWORD cbBytes) {
  // Create a new global memory object
  if (init) {
	 // Determine the number of blocks to allocate
	 HGLOBAL hglbMem;
	 BOOL allocated = FALSE;

	 // Allocate with XMS memory if neither EMS nor CONVENTIONAL are
	 // set, and if we have XMS in the first place
	 if (XMSExists && !(fuFlags & GMEM_CONVENTIONAL) && !(fuFlags & GMEM_EMS)) {
		// Get the number of blocks to allocate (including header)
		WORD numBlocks;
		numBlocks = (WORD) (cbBytes / 1024);
		if ((cbBytes % 1024) != 0)
		  ++numBlocks;
		cbBytes = (DWORD) numBlocks * 1024;

		// Allocate the RAM with XMS
		_DX=numBlocks;
		_AH=9;
		(*XMSFunc)();
		if (_AX==0) {
		  // Error with allocation; return NULL
		  return NULL;
		}
		hglbMem = (HGLOBAL) _DX;

		fuFlags = fuFlags | GMEM_XMS;
		allocated = TRUE;
	 }

	 // Allocate with conventional memory
	 if (!allocated) {
		LPVOID allocation = (LPVOID) farmalloc(cbBytes);

		// If the allocation failed, return NULL
		if (allocation == NULL)
		  return NULL;

		// Make the handle 0x80000000L + the 16-bit real mode address
		// to make a handle that is totally unique for GMEM_CONVENTIONAL
		hglbMem = (HGLOBAL) (0x80000000L + (DWORD) allocation);

		allocated = TRUE;
	 }

	 // If memory was allocated, then initialize it
	 if (allocated)
	 {
		// Add the handle to the handle list and write the header
		if (!AddMemory(hglbMem, cbBytes, fuFlags))
		  return NULL;

		// If GMEM_ZEROINIT was set, then clear the memory
		if (fuFlags & GMEM_ZEROINIT) {
		  LPVOID curBlock;
		  DWORD bytesLeft = cbBytes;
		  DWORD block = 0;
		  while ((bytesLeft / 32767) > 0) {
			 curBlock = GlobalLockEx(hglbMem, TRUE, block, 32767);
			 _fmemset(curBlock, 0, 32767);
			 GlobalUnlock(hglbMem);
			 block += 32767;
			 bytesLeft -= 32767;
		  }
		  if ((bytesLeft) > 0) {
			 curBlock = GlobalLockEx(hglbMem, TRUE, block, bytesLeft);
			 _fmemset(curBlock, 0, (WORD) bytesLeft);
			 GlobalUnlock(hglbMem);
		  }
		  /*DWORD i;
		  LPDWORD mem = (LPDWORD) lockedMemory;
		  LPBYTE memByte;
		  DWORD numBlocks = cbBytes / sizeof(DWORD);

		  // 32-bit aligned clear (fast on 386es and higher)
		  for (i = 0; i < numBlocks; i++) {
			 *mem = 0;
			 mem = (LPDWORD) ((DWORD) mem + sizeof(DWORD));
		  }
		  memByte = (LPBYTE) ((DWORD) mem - 3);

		  // Clear any uncleared bytes
		  for (i = cbBytes - cbBytes % sizeof(DWORD); i < cbBytes; i++) {
			 *memByte = 0;
			 memByte = (LPBYTE) ((DWORD) mem + sizeof(BYTE));
		  }*/
		}

		// Return the desired value
		if (fuFlags & GMEM_FIXED) {
		  // Return the locked memory
		  LPVOID retVal = GlobalLockEx(hglbMem, FALSE, 0, 0);
		  if (!retVal) {
			 // Not enough space or something for whole thing; deallocate and
          // fail function
			 GlobalUnlock(hglbMem);
			 GlobalFree(hglbMem);
			 return NULL;
		  } else
			 return retVal;
		} else {
		  // Return the handle
		  return hglbMem;
		}
	 }

	 // Nothing was allocated!  Return NULL
	 return NULL;
  }

  // KittyRAM not initialized!
  return NULL;
}

extern HGLOBAL GlobalFree(HGLOBAL hglbMem) {
  if (init) {
	 // If the memory object is locked, unlock it
	 DWORD byteOffset;
	 WORD index;
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList(hglbMem, NULL, &byteOffset, &index);

	 if (!lpMemObjList)
		return hglbMem;

	 if (lpMemObjList->locked)
		if (!ForceUnlock(hglbMem))
		  return hglbMem;

	 // Conventional deallocation
	 if ((DWORD) hglbMem & 0x80000000L) {
		farfree((LPVOID)((DWORD) hglbMem - 0x80000000L));
	 }

	 // XMS deallocation
	 if ((DWORD) hglbMem < 65536L) {
		_DX = (WORD) hglbMem;
		_AH = 0x0A;
		(*XMSFunc)();
	 }

	 // Delete the entry from the memory object list

	 // Copy the last one into us, unless if we're the last
	 if (index != nHandles - 1)
		memcpy(&lpMemObjList[index], &lpMemObjList[nHandles - 1], sizeof(MEMOBJ_LIST));

	 // Reduce the size of the list by 1
	 nHandles--;

	 // Shrink the size of the list
	 if (!nHandles) {
		farfree(lpMemObjList);
		lpMemObjList = NULL;
	 } else {
      LPMEMOBJ_LIST newList;
		newList = (LPMEMOBJ_LIST) farrealloc(lpMemObjList, (nHandles) * sizeof(MEMOBJ_LIST));

		// See if the allocation suceeded
		if (!newList)
		  return hglbMem;
		lpMemObjList = newList;
	 }
	 return NULL;
  }

  // Not initialized!
  return hglbMem;
}

// Change the size of the current allocation
extern HGLOBAL GlobalReAlloc(HGLOBAL hglbMem, DWORD cbBytes, DWORD fuFlags) {
  if (init) {
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList(hglbMem, NULL, 0, 0);
	 LPVOID lockedMemory;
	 DWORD oldSize = GlobalSize(hglbMem);

	 if (!lpMemObjList)
		return NULL;

	 // Process the GMEM_MOVEABLE flag
	 if ((fuFlags & GMEM_MOVEABLE) && !(fuFlags & GMEM_MODIFY))
		if (cbBytes == 0)
		  if ((lpMemObjList->locked) || !((lpMemObjList->fuFlags & GMEM_MOVEABLE) && (lpMemObjList->fuFlags & GMEM_DISCARDABLE)))
			 // A special case, if proper flags were not set
			 return NULL;

	 // If the memory object is locked, unlock it
	 if (lpMemObjList->locked)
		if (!ForceUnlock(hglbMem))
		  return NULL;

	 if (!(fuFlags & GMEM_MODIFY)) {
		// Perform the resizing:

		// XMS memory
		if ((DWORD) hglbMem < 65536L) {
		  // Get the number of blocks to allocate (including header)
		  int numBlocks = (int) (cbBytes / 1024);
		  if ((cbBytes % 1024) != 0)
			 ++numBlocks;
		  cbBytes = numBlocks * 1024;

		  // Perform the resizing
		  _BX = numBlocks;
		  _DX = (WORD) hglbMem;
		  _AH = 0x0F;
		  (*XMSFunc)();

		  // If we did not suceeded, then return NULL
		  if (!_AX)
			 return NULL;
		}

		// Conventional resizing
		if ((DWORD) hglbMem & 0x80000000L) {
		  if ((lockedMemory = (LPVOID) farrealloc((LPVOID)((DWORD) hglbMem - 0x80000000L), cbBytes)) == NULL)
			 return NULL;

		  // Set the new handle value
		  hglbMem = (LPVOID) ((DWORD) lockedMemory + 0x80000000L);
		  lpMemObjList->hglbMem = hglbMem;
		}

		// Process the flags, if any
		if ((fuFlags & GMEM_ZEROINIT) && (cbBytes > oldSize)) {
		  // Clear the new space
		  LPVOID curBlock;
		  DWORD bytesLeft = cbBytes - oldSize;
		  DWORD block = oldSize;
		  while ((bytesLeft / 32767) > 0) {
			 curBlock = GlobalLockEx(hglbMem, TRUE, block, 32767);
			 _fmemset(curBlock, 0, 32767);
			 GlobalUnlock(hglbMem);
			 block += 32767;
			 bytesLeft -= 32767;
		  }
		  if ((bytesLeft) > 0) {
			 curBlock = GlobalLockEx(hglbMem, TRUE, block, bytesLeft);
			 _fmemset(curBlock, 0, (WORD) bytesLeft);
			 GlobalUnlock(hglbMem);
		  }
/*		  DWORD i;
		  LPBYTE mem = (LPBYTE) GlobalLock(hglbMem, FALSE, 0, 0);
		  for (i = oldSize; i < cbBytes; i++) {
			 *mem = 0;
			 mem = (LPBYTE) ((DWORD) mem + sizeof(BYTE));
		  }
		  GlobalUnlock(hglbMem);*/
		}

		// Success :-)!  Clean up and return the new memory :-)!
		lpMemObjList->cbBytes = cbBytes;
		return hglbMem;
	 } else {
		// Modify the flags ONLY
		if (fuFlags & GMEM_MOVEABLE) {
		  lpMemObjList->fuFlags = lpMemObjList->fuFlags | GMEM_MOVEABLE;
		  if (fuFlags & GMEM_DISCARDABLE)
			 lpMemObjList->fuFlags = lpMemObjList->fuFlags | GMEM_DISCARDABLE;
		}
		else
		  if ((fuFlags & GMEM_DISCARDABLE) && (lpMemObjList->fuFlags & GMEM_MOVEABLE))
			 lpMemObjList->fuFlags = lpMemObjList->fuFlags | GMEM_DISCARDABLE;
		return hglbMem;
	 }
  }

  // Not initialized
  return NULL;
}

BOOL ForceUnlock(HGLOBAL hglbMem) {
  // Force an unlock by decreasing the lock count all the way
  while (GlobalUnlock(hglbMem));
  return TRUE;
}

extern LPVOID GlobalLockEx(HGLOBAL hglbMem, BOOL conventional, DWORD start, DWORD len) {
  // DWORD start (0) - specifies offset of the beginning of the block to place in conventional if conventional is TRUE
  // DWORD len (0) - specifies length of block to copy to conventional if conventional is true; 0 copies entire block from start to the end of the memory allocation

  // Lock the memory object
  if (init) {
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList(hglbMem, NULL, 0, 0);
	 LPVOID lockedMemory;

	 if (!lpMemObjList)
		return NULL;

	 // Conventional
	 if ((DWORD) hglbMem & 0x80000000L) {
		// Simply return a mathematical modification of the
		// given handle ;-)

		// Increment the lock count
		lpMemObjList->lockedMemory = (LPVOID) ((DWORD) hglbMem - 0x80000000L + start);
		lpMemObjList->locked++;

		return lpMemObjList->lockedMemory;
	 }

	 // XMS
	 if ((DWORD) hglbMem < 65536L) {
		// If we are to not move to conventional, then lock the
		// object and return the address
		if (!conventional) {
/*		  WORD EMBSeg;
		  WORD EMBOff;
		  WORD RetVal;
		  WORD handle = (WORD) hglbMem;*/

		  // Always return NULL -- this part has not been implemented yet
		  return NULL;
/*		  // Enable the A20 line if we need to
		  if (++nA20Locks == 1) {
			 _AH = 0x05;
			 (*XMSFunc)();
		  }

		  if (!_AX)
			 // Error enabling A20!  Return false
			 return NULL;

		  asm {
			 mov	dx, handle
			 mov	ah, 0x0C
			 call	[XMSFunc]
			 mov	RetVal, ax
			 mov	EMBSeg, dx
			 mov	EMBOff, bx
		  }

		  if (!RetVal)
			 return NULL;

		  lockedMemory = (LPVOID) (((DWORD) EMBSeg << 16) + EMBOff);

		  // Lock the XMS block
//		  _AX = 0x0C;
//		  _DX = (WORD) hglbMem;
//		  (*XMSFunc)();

		  // Get the address if it suceeded
//		  lockedMemory = MAKELP(_DX,_BX);

		  // Increment the lock count
		  lpMemObjList->lockedMemory = lockedMemory;
		  lpMemObjList->locked++;

		  return lockedMemory;*/
		} else {
		  // Move the requested block into conventional memory
//		  XMS_MOVEPARAMS xmsMoveParams;
		  DWORD memBytes = lpMemObjList->cbBytes;
		  WORD retVal, xmsMoveParamsOffset;
		  BOOL earlyStart = FALSE;
		  if (len == 0)
			 len = memBytes - start;
		  if ((len % 2) != 0) {
			 // Make the length even (it _must_ be)
			 if (start + len + 1 > memBytes) {
				earlyStart = TRUE;
				start--;
			 } else
				len++;
		  }
		  lockedMemory = (LPVOID) farmalloc(len);

		  // If the memory was not allocated, return NULL
		  if (!lockedMemory)
			 return NULL;

		  // If the A20 line is enabled, then disable it
		  if (nA20Locks > 0) {
			 _AH = 0x06;
			 (*XMSFunc)();
		  }

		  // Do the deed of moving the block :-)
//		  retVal = XMM_MoveMemory(len, (WORD) hglbMem, (DWORD) start, 0, (DWORD) lockedMemory);
		  xmsMoveParams.blockLength = len;
		  xmsMoveParams.sourceHandle = (UINT) hglbMem;
		  xmsMoveParams.sourcePtr = (DWORD) start;
		  xmsMoveParams.destHandle = 0;
		  xmsMoveParams.destPtr = (DWORD) lockedMemory;
		  xmsMoveParamsOffset = FP_OFF(&xmsMoveParams);

		  asm {
			 mov	si, xmsMoveParamsOffset
			 mov	ah, 0x0B
			 call	[XMSFunc]
			 mov	retVal, ax
		  }

/*		  _AH = 0x0B;
		  _SI = FP_OFF(&xmsMoveParams);
		  (*XMSFunc)();*/

		  if (!retVal)
			 // The block was not copied :-(
			 return NULL;

		  // If the A20 line was enabled, then enable it
		  if (nA20Locks > 0) {
			 _AH = 0x05;
			 (*XMSFunc)();
		  }

		  // We have memory :-)!!
		  // Increment the lock count
		  if (earlyStart)
		    lockedMemory = (LPVOID) ((DWORD) lockedMemory + 1);
		  lpMemObjList->lockedMemory = lockedMemory;
		  lpMemObjList->locked++;
		  lpMemObjList->fuFlags = lpMemObjList->fuFlags | GMEM_CONVENTIONAL;
		  lpMemObjList->param1 = start;
		  lpMemObjList->param2 = len;
		  lpMemObjList->param3 = earlyStart;

		  return lockedMemory;
		}
	 }

	 // Invalid handle, no doubt
	 return NULL;
  }

  return NULL;
}

extern BOOL GlobalUnlock(HGLOBAL hglbMem) {
  if (init) {
	 LPVOID lockedMemory;
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList(hglbMem, NULL, 0, 0);

	 // Since we _must_ have lpMemObjList in this function, we must
	 // return FALSE if we don't have it :-(
	 if (!lpMemObjList)
		return FALSE;

	 // Get the locked memory address
	 lockedMemory = lpMemObjList->lockedMemory;

	 // Conventional
	 if ((DWORD) hglbMem & 0x80000000L) {
		// Decrease the lock count
		BOOL locked;
		lpMemObjList->locked--;
		locked = (BOOL) lpMemObjList->locked;
		return (locked)? TRUE: FALSE;
	 }

	 // XMS
	 if ((DWORD) hglbMem < 65536L) {
		if (lpMemObjList->fuFlags & GMEM_CONVENTIONAL) {
		  // This memory was swapped!  Simply move it
		  // back and we'll be on our way :-)
//		  XMS_MOVEPARAMS xmsMoveParams;
		  DWORD start = lpMemObjList->param1;
		  DWORD len = lpMemObjList->param2;
		  WORD xmsMoveParamsOffset, retVal;
		  BOOL earlyStart = lpMemObjList->param3;
		  BYTE errorCode;
		  BOOL locked;
		  lpMemObjList->locked--;
		  locked = (BOOL) lpMemObjList->locked;

		  // Remove the conventional flag
		  lpMemObjList->fuFlags = lpMemObjList->fuFlags - GMEM_CONVENTIONAL;

		  // Adjust for the early start, if any
		  if (earlyStart)
			 lockedMemory = (LPVOID) ((DWORD) lockedMemory - 1);

		  // Do the deed of moving the block :-)
		  xmsMoveParams.blockLength = len;
		  xmsMoveParams.sourceHandle = 0;
		  xmsMoveParams.sourcePtr = (DWORD) lockedMemory;
		  xmsMoveParams.destHandle = (WORD) hglbMem;
		  xmsMoveParams.destPtr = (DWORD) start;
		  xmsMoveParamsOffset = FP_OFF(&xmsMoveParams);

		  asm {
			 mov	si, xmsMoveParamsOffset
			 mov	ah, 0x0B
			 call	[XMSFunc]
			 mov	retVal, ax
			 mov  errorCode, bl
		  }
//		  _AH = 0x0B;
//		  _SI = FP_OFF(&xmsMoveParams);
//		  (*XMSFunc)();

		  if (!retVal)
			 // The block was not copied :-(
			 return errorCode;

		  // Free the conventional memory
		  farfree(lockedMemory);

		  // Finish up here :-)!
		  return (locked)? TRUE: FALSE;
		} else {
		  // Direct-access RAM :-)!
		  WORD handle = (WORD) hglbMem;

		  // Decrease the lock count
		  BOOL locked;
		  lpMemObjList->locked--;
		  locked = (BOOL) lpMemObjList->locked;

		  // Disable the A20 line if we need to
		  if (--nA20Locks == 0) {
			 _AH = 0x06;
			 (*XMSFunc)();
		  }

		  // Unlock the XMS block
		  asm {
			 mov	dx, handle
			 mov	ah, 0x0D
			 call	[XMSFunc]
			 mov	handle, ax
		  }

//		  _AX = 0x0D;
//		  _DX = (WORD) hglbMem;
//		  (*XMSFunc)();

		  // Finish up here :-)!
		  return (locked)? TRUE: FALSE;
		}
	 }
  }

  // Not initialized!
  return NULL;
}

extern DWORD GlobalFlags(HGLOBAL hglbMem) {
  if (init) {
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList(hglbMem, NULL, 0, 0);

	 // If we don't have data on it, we can't do anything :-(
	 if (!lpMemObjList)
		return GMEM_INVALID_HANDLE;

	 // Return the lock count and all that stuff
	 return MAKELONG((lpMemObjList->fuFlags & GMEM_FIXED)? 0: lpMemObjList->locked, lpMemObjList->fuFlags & (GMEM_DDESHARE + GMEM_DISCARDED + GMEM_DISCARDABLE));
  }

  // Not initialized yet
  return GMEM_INVALID_HANDLE;
}

extern HGLOBAL GlobalHandle(LPCVOID lpvMem) {
  if (init) {
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList((HGLOBAL) NULL, (LPVOID) lpvMem, 0, 0);

	 // Get the handle from a pointer we returned
	 if (!lpMemObjList)
		return NULL;

	 return lpMemObjList->hglbMem;
  }

  // Not initialized
  return NULL;
}

extern VOID GlobalMemoryStatus(LPMEMORYSTATUS lpmstMemStat) {
  // Return information on the memory
  lpmstMemStat->dwLength = sizeof(MEMORYSTATUS);
  lpmstMemStat->dwMemoryLoad = 1;
  lpmstMemStat->dwTotalPhys = 1;
  lpmstMemStat->dwAvailPhys = 1;
  lpmstMemStat->dwTotalPageFile = 0;
  lpmstMemStat->dwAvailPageFile = 0;
  lpmstMemStat->dwTotalVirtual = 0;
  lpmstMemStat->dwAvailVirtual = 0;
  lpmstMemStat->fMemTypes = ((XMSExists)? GMEM_XMS: 0) | ((EMSExists)? GMEM_EMS: 0);
}

extern DWORD GlobalSize(HGLOBAL hglbMem) {
  if (init) {
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList(hglbMem, NULL, 0, 0);

	 // If we don't have data on it, we can't do anything :-(
	 if (!lpMemObjList)
		return GMEM_INVALID_HANDLE;

	 // Return the lock count and all that stuff
	 return lpMemObjList->cbBytes;
  }

  // Not initialized yet
  return 0;
}

extern BOOL GlobalStart(VOID) {
  // Start a KittyRAM session

  // Check for XMS
  _AX=0x4300;
  geninterrupt(0x2F);
  XMSExists = (_AL==0x80);

  // If we have XMS, then get the entry point
  if (XMSExists) {
	 _AX=0x4310;
	 geninterrupt(0x2F);
	 XMSFunc= (void (far *)(void)) MAKELP(_ES,_BX);
  }

  // We don't support EMS right now :-(
  EMSExists = FALSE;

  // Initialize the memObjList
  init = TRUE;
  lpMemObjList = NULL;
  nHandles = 0;

  // Set the number of A20 locks to zero
  nA20Locks = 0;

  return TRUE;
}

extern DWORD GlobalEnd(VOID){
  // End a KittyRAM session

  // Go through the entire memory object list, backwards, freeing
  // everything in its path!
  WORD i;
  WORD nFreed = nHandles;

  // Free everything, including the memory object list
  for (i = nHandles; i > 0; i--)
	 GlobalFree(lpMemObjList[i - 1].hglbMem);

  // And we are done :-)!
  init = FALSE;
  return nFreed;
}

extern BOOL GlobalReady(VOID) {
  // Have we been initialized yet?
  return init;
}

extern BOOL GlobalArrayStart(HGLOBAL hglbMem, DWORD dwElementSize, DWORD nElements, DWORD nCachedElements) {
  if (init) {
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList(hglbMem, NULL, 0, 0);

	 // If we don't have data on it, we can't do anything :-(
	 if (!lpMemObjList)
		return FALSE;

	 // If the array was locked, then unlock it
	 if (lpMemObjList->arrayLocked)
		GlobalUnlock(hglbMem);

	 // Initialize members
	 lpMemObjList->dwElementSize = dwElementSize;
	 lpMemObjList->nElements = nElements;
	 lpMemObjList->curElement = 0;
	 lpMemObjList->arrayLocked = FALSE;
	 lpMemObjList->nCachedElements = nCachedElements;

	 return TRUE;
  }

  // Not initialized yet
  return FALSE;
}

extern LPVOID GlobalArrayFirst(HGLOBAL hglbMem) {
  if (init) {
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList(hglbMem, NULL, 0, 0);
	 LPVOID lockedMemory;
	 if (!lpMemObjList)
		return NULL;

	 // Check to see if this beyond the last element
	 if (lpMemObjList->curElement > lpMemObjList->nElements)
		return NULL;

	 // See if the element is in the cache
	 if (lpMemObjList->curElement / lpMemObjList->nCachedElements == 0) {
		// We are currently in the same block; return the memory
		lpMemObjList->curElement = 0;
		return lpMemObjList->lockedMemory;
	 }

	 // Get the first element in the array
	 if (lpMemObjList->arrayLocked) {
		if (lpMemObjList->curElement != 0) {
		  GlobalUnlock(hglbMem);
		  if ((DWORD) (lockedMemory = GlobalLockEx(hglbMem, TRUE, 0, lpMemObjList->dwElementSize * lpMemObjList->nCachedElements)) == NULL)
			 lpMemObjList->arrayLocked = FALSE;
		  return lockedMemory;
		} else
		  return lpMemObjList->lockedMemory;
	 } else
		if ((lockedMemory = GlobalLockEx(hglbMem, TRUE, 0, lpMemObjList->dwElementSize * lpMemObjList->nCachedElements)) != NULL)
		  lpMemObjList->arrayLocked = TRUE;
		return lockedMemory;
  }
  return NULL;
}

extern LPVOID GlobalArrayNext(HGLOBAL hglbMem) {
  if (init) {
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList(hglbMem, NULL, 0, 0);
	 LPVOID lockedMemory;
	 DWORD origPos;
	 if (!lpMemObjList)
		return NULL;

	 // Check to see if this beyond the last element if this isn't the first call
	 origPos = lpMemObjList->curElement;
	 if (lpMemObjList->arrayLocked)
		lpMemObjList->curElement += 1;
	 if (lpMemObjList->curElement > lpMemObjList->nElements) {
		if (origPos < lpMemObjList->curElement)
		  lpMemObjList->curElement -= 1;
		return NULL;
	 }

	 // See if the element is in the cache
	 if (origPos / lpMemObjList->nCachedElements == lpMemObjList->curElement / lpMemObjList->nCachedElements) {
		// We are currently in the same block; return the memory
		return (LPVOID) ((DWORD) lpMemObjList->lockedMemory + (lpMemObjList->curElement % lpMemObjList->nCachedElements) * lpMemObjList->dwElementSize);
	 }

	 // Get the next element in the array
	 if (lpMemObjList->arrayLocked)
		GlobalUnlock(hglbMem);
	 if ((DWORD) (lockedMemory = GlobalLockEx(hglbMem, TRUE, lpMemObjList->dwElementSize * lpMemObjList->nCachedElements * (lpMemObjList->curElement / lpMemObjList->nCachedElements), lpMemObjList->dwElementSize * lpMemObjList->nCachedElements)) == NULL)
		lpMemObjList->arrayLocked = FALSE;
	 else {
		lpMemObjList->arrayLocked = TRUE;
		lockedMemory = (LPVOID) ((DWORD) lockedMemory + (lpMemObjList->curElement % lpMemObjList->nCachedElements) * lpMemObjList->dwElementSize);
	 }
	 return lockedMemory;
  }
  return NULL;
}

extern LPVOID GlobalArrayElement(HGLOBAL hglbMem, DWORD dwElement) {
  if (init) {
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList(hglbMem, NULL, 0, 0);
	 LPVOID lockedMemory;
	 if (!lpMemObjList)
		return NULL;

	 // Check to see if this beyond the last element
	 if (dwElement > lpMemObjList->nElements)
		return NULL;

	 // See if the element is in the cache
	 if (lpMemObjList->curElement / lpMemObjList->nCachedElements == dwElement / lpMemObjList->nCachedElements) {
		// We are currently in the same block; return the memory
		lpMemObjList->curElement = dwElement;
		return (LPVOID) ((DWORD) lpMemObjList->lockedMemory + (lpMemObjList->curElement % lpMemObjList->nCachedElements) * lpMemObjList->dwElementSize);
	 }

	 // Get the specified element in the array
	 if (lpMemObjList->arrayLocked)
		if (lpMemObjList->curElement == dwElement)
		  return lpMemObjList->lockedMemory;
		else
		  GlobalUnlock(hglbMem);

	 if ((DWORD) (lockedMemory = GlobalLockEx(hglbMem, TRUE, lpMemObjList->dwElementSize * lpMemObjList->nCachedElements * (dwElement / lpMemObjList->nCachedElements), lpMemObjList->dwElementSize * lpMemObjList->nCachedElements)) == NULL)
		lpMemObjList->arrayLocked = FALSE;
	 else {
		lpMemObjList->curElement = dwElement;
		lpMemObjList->arrayLocked = TRUE;
		lockedMemory = (LPVOID) ((DWORD) lockedMemory + (dwElement % lpMemObjList->nCachedElements) * lpMemObjList->dwElementSize);
	 }
	 return lockedMemory;
  }
  return NULL;
}

extern BOOL GlobalArrayEnd(HGLOBAL hglbMem) {
  if (init) {
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList(hglbMem, NULL, 0, 0);
	 if (!lpMemObjList)
		return NULL;

	 // Unlock the array lock
	 if (lpMemObjList->arrayLocked) {
		GlobalUnlock(hglbMem);
		lpMemObjList->arrayLocked = FALSE;
	 }
	 return TRUE;
  }
  return FALSE;
}

extern VOID ZeroMemory(LPVOID Destination, DWORD Length) {
  if (init) {
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList(NULL, Destination, 0, 0);
	 if (lpMemObjList) {
		// This memory happens to be one of our global objects; clear it
		// with the whole allocation in mind; not just the window
		LPVOID curBlock;
		DWORD bytesLeft = Length;
		DWORD block = 0;
		while ((bytesLeft / 32767) > 0) {
		  curBlock = GlobalLockEx(lpMemObjList->hglbMem, TRUE, block, 32767);
		  _fmemset(curBlock, 0, 32767);
		  GlobalUnlock(lpMemObjList->hglbMem);
		  block += 32767;
		  bytesLeft -= 32767;
		}
		if ((bytesLeft) > 0) {
		  curBlock = GlobalLockEx(lpMemObjList->hglbMem, TRUE, block, bytesLeft);
		  _fmemset(curBlock, 0, (WORD) bytesLeft);
		  GlobalUnlock(lpMemObjList->hglbMem);
		}
		return;
	 }
  }

  // Clear the standard amount
  _fmemset(Destination, 0, (WORD) Length);
}

extern VOID FillMemory(LPVOID Destination, DWORD Length, BYTE Fill) {
  if (init) {
	 LPMEMOBJ_LIST lpMemObjList = GetMemObjList(NULL, Destination, 0, 0);
	 if (lpMemObjList) {
		// This memory happens to be one of our global objects; clear it
		// with the whole allocation in mind; not just the window
		LPVOID curBlock;
		DWORD bytesLeft = Length;
		DWORD block = 0;
		while ((bytesLeft / 32767) > 0) {
		  curBlock = GlobalLockEx(lpMemObjList->hglbMem, TRUE, block, 32767);
		  _fmemset(curBlock, Fill, 32767);
		  GlobalUnlock(lpMemObjList->hglbMem);
		  block += 32767;
		  bytesLeft -= 32767;
		}
		if ((bytesLeft) > 0) {
		  curBlock = GlobalLockEx(lpMemObjList->hglbMem, TRUE, block, bytesLeft);
		  _fmemset(curBlock, Fill, (WORD) bytesLeft);
		  GlobalUnlock(lpMemObjList->hglbMem);
		}
		return;
	 }
  }

  // Clear the standard amount
  _fmemset(Destination, Fill, (WORD) Length);
}

