#include "winmem.h"


/* Global Memory Flags */
#define GMEM_FIXED          0x0000
#define GMEM_XMS            0x0001 // Object is in XMS
#define GMEM_MOVEABLE       0x0002
#define GMEM_EMS				 0x0004 // Object is in EMS
#define GMEM_CONVENTIONAL   0x0008 // Object is in conventional;
#define GMEM_NOCOMPACT      0x0010   // if GMEM_CONVENTIONAL is set
#define GMEM_NODISCARD      0x0020   // and GMEM_?MS is set, the data
#define GMEM_ZEROINIT       0x0040   // is temporarily in conventional;
#define GMEM_MODIFY         0x0080   // otherwise, memory is set in
#define GMEM_DISCARDABLE    0x0100   // conventional with malloc
#define GMEM_NOT_BANKED     0x1000
#define GMEM_SHARE          0x2000
#define GMEM_DDESHARE       0x2000
#define GMEM_NOTIFY         0x4000
#define GMEM_LOWER          GMEM_NOT_BANKED
#define GMEM_VALID_FLAGS    0x7F7F
#define GMEM_INVALID_HANDLE 0x8000

#define GHND                (GMEM_MOVEABLE | GMEM_ZEROINIT)
#define GPTR                (GMEM_FIXED | GMEM_ZEROINIT)

#define GlobalLRUNewest( h )    (HANDLE)(h)
#define GlobalLRUOldest( h )    (HANDLE)(h)
#define GlobalDiscard( h )      GlobalReAlloc( (h), 0, GMEM_MOVEABLE )

/* Flags returned by GlobalFlags (in addition to GMEM_DISCARDABLE) */
#define GMEM_DISCARDED      0x4000
#define GMEM_LOCKCOUNT      0x00FF

typedef struct _MEMORYSTATUS {
	 DWORD dwLength;
	 DWORD dwMemoryLoad;
	 DWORD dwTotalPhys;
	 DWORD dwAvailPhys;
	 DWORD dwTotalPageFile;
	 DWORD dwAvailPageFile;
	 DWORD dwTotalVirtual;
	 DWORD dwAvailVirtual;
	 DWORD fMemTypes; // includes the types of memory in the system
} MEMORYSTATUS, *LPMEMORYSTATUS;


// Functions

extern HGLOBAL GlobalAlloc(DWORD fuFlags, DWORD cbBytes);
extern HGLOBAL GlobalFree(HGLOBAL hglbMem);
extern HGLOBAL GlobalReAlloc(HGLOBAL hglbMem, DWORD cbBytes, DWORD fuFlags);
extern LPVOID GlobalLockEx(HGLOBAL hglbMem, BOOL conventional, DWORD start, DWORD len);  // DWORD start (0) - specifies offset of the beginning of the block to place in conventional if conventional is TRUE
																				 // DWORD len (0) - specifies length of block to copy to conventional if conventional is true; 0 copies entire block from start to the end of the memory allocation
#define GlobalLock(hglbMem)   (LPVOID) GlobalLockEx(hglbMem, TRUE, 0, 0)
extern DWORD GlobalFlags(HGLOBAL hglbMem);
extern HGLOBAL GlobalHandle(LPCVOID lpvMem);
extern BOOL GlobalUnlock(HGLOBAL hglbMem);
extern VOID GlobalMemoryStatus(LPMEMORYSTATUS lpmstMemStat);
extern DWORD GlobalSize(HGLOBAL hglbMem);
extern BOOL GlobalStart(VOID); // Start a KittyRAM session
extern DWORD GlobalEnd(VOID);  // End a KittyRAM session
extern BOOL GlobalReady(VOID);
extern BOOL GlobalArrayStart(HGLOBAL hglbMem, DWORD dwElementSize, DWORD nElements);
extern LPVOID GlobalArrayFirst(HGLOBAL hglbMem);
extern LPVOID GlobalArrayNext(HGLOBAL hglbMem);
extern LPVOID GlobalArrayElement(HGLOBAL hglbMem, DWORD dwElement);
extern BOOL GlobalArrayEnd(HGLOBAL hglbMem);
extern VOID ZeroMemory(LPVOID Destination, DWORD Length);
extern VOID FillMemory(LPVOID Destination, DWORD Length, BYTE Fill);

